--[[

̃t@CRampageAÏꕔłB

RampageAI̓t[\tgł; Free Software FoundationzGNUʌOp_񏑂̏̂ƂōĔzzщς邱Ƃł܂;
GNUCZXo[W2A܂(Ȃ̑I)2ȑÕo[WɂāB

RampageAÍALpłƂ]Ĕzz܂B
*ۏ؂Ȃ*; s̐A̖ړÎ߂̓K̈ÖٓIȕۏ؂ȂĂB 
̑̏ڍׂɊւẮAGNUʌOp_񏑂ĂB

RampageAIƋGNUʌOp_񏑂̃Rs[𓯎Ɏ󂯎ׂł;
łȂȂAȉ̏֏ȂB
Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

----------------------
-- Attack Functions --
----------------------

-- Simple attack function
function AI_Attack(actor)
	Attack(HomunID,actor)
end

-- Attack (with skills too) function
function AI_SkilledAttack(actor)
	-- Send a skill attack
	SendAttackSkill(actor)
	
	-- After everything else, try to do a normal attack
	Attack(HomunID,actor)
end

-- Send an attack skill command
function SendAttackSkill(actor)
	-- Check if skills are allowed
	if not MSkillsAllowed(actor) then
		-- If they aren't allowed, return false
		return false
	end

	-- Get the range
	local range = AADistance(HomunID, actor)

	-- Get a skill to use
	local skill,level
	
	-- Check the type
	if range <= HomunRange then
		skill,level = CheckAttackSkill(SKILL_TYPE_MELEE)
	else
		skill,level = CheckAttackSkill(SKILL_TYPE_RANGED,range)
	end

	-- Check if we got a skill
	if skill ~= 0 then
		-- Use the skill
		UseAutoSkill(level,skill,actor)
		
		return true
	end

	-- If we didn't use a skill, return false
	return false
end

-- Function to check if we should use a skill
function AI_CheckAttackSkill(type,range)
	--[[
		If "type" is melee, it'll check melee and ranged skills,
		otherwise the only other type that should be used is
		ranged, and it will only check those
	--]]

	-- First things first, check if its allowed
	if not SkillAllowed(type) then
		-- No skill, sorry
		return 0,0
	end

	-- Grab the amount of SP that we're allowed to use
	local sp = SPUsable(HomunID)

	-- Check if we should do melee skills as well
	local i,skill,j
	if type == SKILL_TYPE_MELEE and Table.getn(AttackSkills) > 0 then
		-- Loop through all our melee skills
		for i,skill in ipairs(AttackSkills) do
			-- Loop through the SP table
			for j=5,MinAttackSkillLevel,-1 do
				-- If we have enough SP for it, use it
				if sp >= skill.sp[j] then
					-- Use the skill
					return skill.id,j
				end
			end
		end
	end

	-- Check if we have any ranged skills to try and use
	if Table.getn(RangedSkills) < 1 then
		-- We don't, so no skill this time
		return 0,0
	end

	-- Check if range matters
	if range == nil then
		-- Doesn't matter, set it super high
		range = 100
	end

	-- Now check for a ranged skill to use
	for i,skill in ipairs(RangedSkills) do
		-- Check if the skill is in range
		if skill.range <= range then
			-- Loop through the SP table
			for j=5,MinAttackSkillLevel,-1 do
				-- Check its SP
				if sp >= skill.sp[j] then
					-- We got our skill, return it
					return skill.id,j
				end
			end
		end
	end

	-- No valid skills, return 0,0
	return 0,0
end

-- Function to use a skill
function AI_UseSkill(level,type,target)
	-- Check if its time to use a skill
	if LastAttackSkill == 0 or
		GetTick() >= (MSkillTimeout(target) + LastAttackSkill) then
		-- Attack-skill usage is alright now

		-- Log it
		Log("Automatically using skill #%d (level %d) against actor #%d",
			type,level,target)

		-- Use the skill
		SkillObject(HomunID,level,type,target)

		-- Reset the timeout
		LastAttackSkill = GetTick()
	end
end

------------------------
-- Target Acquisition --
------------------------

-- Function to acquire a target
function AcquireTarget()
	-- Check if we have a target-lock
	if Target.Lock then
		-- Reset the swap counter
		Target.Swap = 0
		
		-- Don't try to acquire another target
		return
	end
	
	-- Determine which function to use to acquire targets
	local AcquireTargetFunc = PassiveAcquire
	if Mode == MODE_AGGRO then
		AcquireTargetFunc = AggroAcquire
	end

	-- Save our previous target
	local saved = Target.Previous

	-- Acquire a target, and check it
	if AcquireTargetFunc() == -1 then
		-- No target, so return right away
		return
	end

	-- Hold position is useless now
	HoldPosition = false

	-- Ensure we have an attack command on the AI_seq
	if not InSequence(SEQ_ATTACK) then
		List.pushright(AI_seq,{ SEQ_ATTACK })
	end

	-- Check if we have a different target
	if Target.ID ~= Target.Previous then
		-- Log our new target
		Log("Now targetting actor #%d",Target.ID)
	end

	-- Check if our target is the previous one
	if saved == Target.ID and
		saved ~= Target.Previous then
		-- Seems like we swapped targets

		-- Increment our swap counter
		Target.Swap = Target.Swap + 1

		-- Check if we've swapped too often
		if Target.Swap == 3 then
			-- Log it
			Log("Locking on to actor #%d, due to too many target swaps",
				Target.ID)
			
			-- Set target lock
			Target.Lock = true
		end
	else
		-- Reset our swap counter
		Target.Swap = 0
	end
end

-- Determine which enemies are attacking us
function AI_PassiveAcquire(reused)
	-- Identify enemies attacking our owner
	local ownerAttackers = MPrioritySieve(IgnoredSieve(ActorEnemies(OwnerID)))

	-- Identify enemies attacking the homunculus
	local homunAttackers = MPrioritySieve(IgnoredSieve(ActorEnemies(HomunID)))

	-- Check if we care about our party members
	local partyAttackers = Table.new()
	if DefendParty then
		-- Loop through each party member
		local i,partyactor
		for i,partyactor in ipairs(PartyMembers) do
			-- Identify enemies attacking them
			partyAttackers = ActorEnemies(partyactor,partyAttackers)
		end

		-- Remove ignored monsters, and grab the table with the highest priority
		partyAttackers = MPrioritySieve(IgnoredSieve(partyAttackers))
	end

	-- Default to prioritizing monsters on homunculus
	local ownerDefend = false

	-- Check the number of attackers on our owner
	if PrioritizeOwnerDefense <= Table.getn(ownerAttackers) or
		Table.getn(homunAttackers) < 1 then
		-- Prioritize defense of our owner...

		-- Unless homunculus has too many attackers
		if PrioritizeHomunDefense > Table.getn(homunAttackers) then
			-- Owner needs help first
			ownerDefend = true
		end
	end

	-- Check if we should defend our owner first
	if ownerDefend then
		-- Defend owner first
		target = ATNearest(HomunID,ownerAttackers)
	else
		-- Defend homunculus first
		target = ATNearest(HomunID,homunAttackers)
	end

	-- Check if we found a target
	if target == -1 and DefendParty then
		-- Check if we can defend a party member
		target = ATNearest(HomunID,partyAttackers)
	end

	-- Check if we should update the Target structure
	if reused ~= true then
		Target.Previous = Target.ID
		Target.ID = target
	end

	return target
end

-- Determine the closest target to attack
function AI_AggroAcquire(reused)
	-- First and foremost, defend ourselves
	local target = PassiveAcquire(true)

	-- Check if we got a target thats attacking us
	if target == -1 then
		-- Loop through all actors, determining which are enemies
		local enemies = Table.new()
		local i,actor,motion
		for actor,i in pairs(ActorTargets) do
			-- Make sure the actor isn't homunculus or owner
			if actor ~= OwnerID and actor ~= HomunID then
				-- Get the motion of the actor
				motion = GetV(V_MOTION,actor)

				-- Make sure the actor is on the screen and not dead
				if motion ~= MOTION_DEAD and ActorIsInSight(actor) then
					-- Check if its a valid monster (and its allowed)
					if IsMonster(actor) == 1 and
						MonsterAllowed(actor) and
						not IsIgnored(actor) then
						--[[
				Note: PVP players appear as monsters unless they're
					in your party (this is actually not true!)
						--]]

						-- Check if we'll "kill-steal"
						if not KillSteal(actor) then
							-- Add the monster to the table
							Table.insert(enemies,actor)
						end
					end
				end
			end
		end

		-- Find the nearest enemy, from the list of highest priority
		target = ATNearest(HomunID,MPrioritySieve(enemies))
	end

	-- Check if we should update the Target structure
	if reused ~= true then
		Target.Previous = Target.ID
		Target.ID = target
	end

	return target
end

-----------------------
-- Support Functions --
-----------------------

-- Check if we would kill steal by attacking the actor
function KillSteal(actor)
--[[
	-- Check if we're in a PVP zone
	if PVPZone then
		-- Can't kill-steal in a PVP zone

		-- And never check again
		KillSteal = FalseReturn

		return false
	end
--]]

	-- Check if its a monster
	if IsMonster(actor) ~= 1 then
		-- Can't kill-steal a player O.o
		return false
	end

	-- Check if the monster is targetted and locked
	if actor == Target.ID and Target.Lock then
		-- It might be kill stealing, but we won't count it
		return false
	end

	-- Get the monster's target
	local target = GetV(V_TARGET,actor)

	-- Check if the monster is targetting something
	if target ~= 0 then
		-- Check if the monster is targetting homunculus or owner
		if target == HomunID or target == OwnerID then
			-- Won't be "kill-stealing"
			return false
		end

		-- Check if monster is targetting a party member
		if DefendParty and IsPartyMember(target) then
			-- We are supposed to defend party members,
			-- so it isn't kill-stealing
			return false
		end

		-- Check if monster is targetting another monster
		-- (with a heal spell or such)
		if IsMonster(target) == 1 then
			-- Can't kill-steal from monsters
			return false
		end

		-- It would be kill stealing
		return true
	end

	-- Check if any non-monsters are targetting it
	local table = ActorTargets[actor]
	if table ~= nil then
		if Table.getn(table) > 0 then
			-- Loop through, trying to find a non-monster
			local i,actor
			for i,actor in ipairs(table) do
				if actor ~= HomunID and
					actor ~= OwnerID and
					IsMonster(actor) ~= 1 and
					not IsPartyMember(actor) then
					-- Its not a monster or a party member,
					-- so likely we'd be kill stealing
					return true
				end
			end
		end
	end

	-- It won't skill steal
	return false
end
